#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022 Salvador E. Tropea
# Copyright (c) 2020-2022 Instituto Nacional de Tecnologïa Industrial
# Copyright (c) 2019 Jesse Vincent (@obra)
# Copyright (c) 2018-2019 Seppe Stas (@seppestas) (Productize SPRL)
# Copyright (c) 2015-2016 Scott Bezek (@scottbez1)
# License: Apache 2.0
# Project: KiAuto (formerly kicad-automation-scripts)
# Adapted from: https://github.com/obra/kicad-automation-scripts
"""
Utility functions for UI automation with xdotool in a virtual framebuffer
with XVFB. Also includes utilities for accessing the clipboard for easily
and efficiently copy-pasting strings in the UI.

Based on splitflap/electronics/scripts/export_util.py by Scott Bezek
"""
import re
import os
import sys
from subprocess import Popen, CalledProcessError, TimeoutExpired, call, check_output, STDOUT, DEVNULL, run, PIPE
import time
import shutil
import signal
from tempfile import mkdtemp
from contextlib import contextmanager
# python3-xvfbwrapper
from xvfbwrapper import Xvfb
from kiauto.file_util import get_log_files
from kiauto.misc import KICAD_VERSION_5_99, MISSING_TOOL, KICAD_DIED

from kiauto import log
logger = log.get_logger(__name__)
time_out_scale = 1.0
img_tmp_dir = None


def set_time_out_scale(scale):
    global time_out_scale
    time_out_scale = scale


class PopenContext(Popen):

    def __exit__(self, type, value, traceback):
        logger.debug("Closing pipe with %d", self.pid)
        # Note: currently we don't communicate with the child so these cases are never used.
        # I keep them in case they are needed, but excluded from the coverage.
        # Also note that closing stdin needs extra handling, implemented in the parent class
        # but not here.
        if self.stdout:
            self.stdout.close()  # pragma: no cover
        if self.stderr:
            self.stderr.close()  # pragma: no cover
        if self.stdin:
            self.stdin.close()   # pragma: no cover
        if type:
            logger.debug("Terminating %d", self.pid)
            # KiCad nightly uses a shell script as intermediate to run setup the environment
            # and run the proper binary. If we simply call "terminate" we just kill the
            # shell script. So we create a group and then kill the whole group.
            try:
                os.killpg(os.getpgid(self.pid), signal.SIGTERM)
            except ProcessLookupError:
                pass
            # self.terminate()
        # Wait for the process to terminate, to avoid zombies.
        try:
            # Wait for 3 seconds
            self.wait(3)
            retry = False
        except TimeoutExpired:  # pragma: no cover
            # The process still alive after 3 seconds
            retry = True
            pass
        if retry:  # pragma: no cover
            logger.debug("Killing %d", self.pid)
            # We shouldn't get here. Kill the process and wait upto 10 seconds
            os.killpg(os.getpgid(self.pid), signal.SIGKILL)
            # self.kill()
            self.wait(10)


def wait_xserver(out_dir, num_try):
    global time_out_scale
    timeout = 10*time_out_scale
    DELAY = 0.5
    logger.debug('Waiting for virtual X server ...')
    try:
        logger.debug('Current DISPLAY is '+os.environ['DISPLAY'])
    except KeyError:
        logger.error('Missing DISPLAY on wait_xserver!')
    if shutil.which('setxkbmap'):
        cmd = ['setxkbmap', '-query']
    elif shutil.which('xset'):  # pragma: no cover
        cmd = ['xset', 'q']
    else:  # pragma: no cover
        cmd = ['ls']
        logger.warning('No setxkbmap nor xset available, unable to verify if X is running')
    for i in range(int(timeout/DELAY)):
        logger.debug('Checking using '+str(cmd))
        flog_out, flog_err = get_log_files(out_dir, cmd[0])
        ret = call(cmd, stdout=flog_out, stderr=flog_err, close_fds=True)
        if not ret:
            time_wait = 0.5*(num_try+1)*(num_try+1)*time_out_scale
            if log.get_level() > 2:
                logger.debug(str(cmd)+' returned 0')
                logger.debug('Waiting {} seconds before using the X server'.format(time_wait))
            # On GitLab I saw setxkbmap success and recordmydesktop and KiCad failing to connect to X
            # One possible solution is a wait here.
            # Another is detecting KiCad exited.
            time.sleep(time_wait)
            return
        logger.debug('   Retry')
        time.sleep(DELAY)
    raise RuntimeError('Timed out waiting for virtual X server')


def wait_wm():
    global time_out_scale
    timeout = 10*time_out_scale
    DELAY = 0.5
    logger.debug('Waiting for Window Manager ...')
    if shutil.which('wmctrl'):
        cmd = ['wmctrl', '-m']
    else:  # pragma: no cover
        logger.warning('No wmctrl, unable to verify if WM is running')
        time.sleep(2)
        return
    logger.debug('Checking using '+str(cmd))
    for i in range(int(timeout/DELAY)):
        ret = call(cmd, stdout=DEVNULL, stderr=STDOUT, close_fds=True)
        if not ret:
            return
        logger.debug('   Retry')
        time.sleep(DELAY)
    raise RuntimeError('Timed out waiting for WM server')


@contextmanager
def start_wm(do_it):
    if do_it:
        cmd = ['fluxbox']
        logger.debug('Starting WM: '+str(cmd))
        with PopenContext(cmd, stdout=DEVNULL, stderr=DEVNULL, close_fds=True, start_new_session=True) as wm_proc:
            wait_wm()
            try:
                yield
            finally:
                logger.debug('Terminating the WM')
                # Fluxbox sometimes will ignore SIGTERM, we can just kill it
                wm_proc.kill()
    else:
        yield


@contextmanager
def start_record(do_record, video_dir, video_name):
    if do_record:
        video_filename = os.path.join(video_dir, video_name)
        cmd = ['recordmydesktop', '--overwrite', '--no-sound', '--no-frame', '--on-the-fly-encoding',
               '-o', video_filename]
        logger.debug('Recording session with: '+str(cmd))
        flog_out, flog_err = get_log_files(video_dir, cmd[0])
        with PopenContext(cmd, stdout=flog_out, stderr=flog_err, close_fds=True, start_new_session=True) as screencast_proc:
            try:
                yield
            finally:
                logger.debug('Terminating the session recorder')
                screencast_proc.terminate()
    else:
        yield


@contextmanager
def start_x11vnc(do_it, old_display):
    if do_it:
        if not shutil.which('x11vnc'):
            logger.error("x11vnc isn't installed, please install it")
            yield
        else:
            cmd = ['x11vnc', '-display', os.environ['DISPLAY'], '-localhost']
            logger.debug('Starting VNC server: '+str(cmd))
            with PopenContext(cmd, stdout=DEVNULL, stderr=DEVNULL, close_fds=True, start_new_session=True) as x11vnc_proc:
                if old_display is None:
                    old_display = ':0'
                logger.debug('To monitor the Xvfb now you can start: "ssvncviewer '+old_display+'"(or similar)')
                try:
                    yield
                finally:
                    logger.debug('Terminating the x11vnc server')
                    x11vnc_proc.terminate()
    else:
        yield


@contextmanager
def recorded_xvfb(cfg, num_try=0):
    try:
        old_display = os.environ['DISPLAY']
    except KeyError:
        old_display = None
    with Xvfb(width=cfg.rec_width, height=cfg.rec_height, colordepth=cfg.colordepth):
        wait_xserver(cfg.output_dir, num_try)
        with start_x11vnc(cfg.start_x11vnc, old_display):
            with start_wm(cfg.use_wm):
                with start_record(cfg.record, cfg.video_dir, cfg.video_name):
                    yield


def xdotool(command, id=None):
    if id is not None:
        command.insert(1, str(id))
        command.insert(1, '--window')
    # logger.debug(['xdotool'] + command)
    return check_output(['xdotool'] + command, stderr=DEVNULL).decode()
    # return check_output(['xdotool'] + command)


# def clipboard_store(string):
#     # I don't know how to use Popen/run to make it run with pipes without
#     # either blocking or losing the messages.
#     # Using files works really well.
#     logger.debug('Clipboard store "'+string+'"')
#     # Write the text to a file
#     fd_in, temp_in = tempfile.mkstemp(text=True)
#     os.write(fd_in, string.encode())
#     os.close(fd_in)
#     # Capture output
#     fd_out, temp_out = tempfile.mkstemp(text=True)
#     process = Popen(['xclip', '-selection', 'clipboard', temp_in], stdout=fd_out, stderr=STDOUT)
#     ret_code = process.wait()
#     os.remove(temp_in)
#     os.lseek(fd_out, 0, os.SEEK_SET)
#     ret_text = os.read(fd_out, 1000)
#     os.close(fd_out)
#     os.remove(temp_out)
#     ret_text = ret_text.decode()
#     if ret_text:  # pragma: no cover
#         logger.error('Failed to store string in clipboard')
#         logger.error(ret_text)
#         raise
#     if ret_code:  # pragma: no cover
#         logger.error('Failed to store string in clipboard')
#         logger.error('xclip returned %d' % ret_code)
#         raise


def text_replace(string):
    """ Used to replace a text in an input text widget. """
    delay = str(int(12*time_out_scale))
    cmd = ['key', '--delay', delay, 'ctrl+a', 'type', '--delay', delay, string]
    logger.debug('text_replace with: {}'.format(cmd))
    xdotool(cmd)


def clipboard_retrieve():
    p = Popen(['xclip', '-o', '-selection', 'clipboard'], stdout=PIPE, stderr=STDOUT)
    output = ''
    for line in p.stdout:
        output += line.decode()
    logger.debug('Clipboard retrieve "'+output+'"')
    return output


def get_windows(all=False):
    cmd = ['search', '--name', '.*']
    if not all:
        cmd.insert(1, '--onlyvisible')
    ids = xdotool(cmd).splitlines()
    res = []
    for i in ids:
        name = ''
        try:
            name = xdotool(['getwindowname', i])[:-1]
        except CalledProcessError:
            name = '** No longer there **'
        res.append((i, name))
        if log.get_level() >= 2:
            logger.debug('get_windows {} {}'.format(i, name))
    if log.get_level() >= 2:
        logger.debug('get_windows end of list')
    return res


def debug_window(id=None):  # pragma: no cover
    if log.get_level() < 2:
        return
    if shutil.which('xprop'):
        if id is None:
            try:
                id = xdotool(['getwindowfocus']).rstrip()
            except CalledProcessError:
                logger.debug('xdotool getwindowfocus failed!')
                pass
        if id:
            call(['xprop', '-id', id])
    if shutil.which('vmstat'):
        call(['vmstat', '-s'])
    if shutil.which('uptime'):
        call(['uptime'])
    logger.debug("Visible windows:")
    for i in get_windows():
        logger.debug("Window ID: `{}` ; name: `{}`".format(i[0], i[1]))
    logger.debug("All windows:")
    for i in get_windows(all=True):
        logger.debug("Window ID: `{}` ; name: `{}`".format(i[0], i[1]))


def wait_focused(id, timeout=10):
    global time_out_scale
    timeout *= time_out_scale
    DELAY = 0.5
    logger.debug('Waiting for %s window to get focus...', id)
    for i in range(int(timeout/DELAY)):
        cur_id = xdotool(['getwindowfocus']).rstrip()
        logger.debug('Currently focused id: %s', cur_id)
        if cur_id == id:
            return
        time.sleep(DELAY)
    debug_window(cur_id)  # pragma: no cover
    raise RuntimeError('Timed out waiting for %s window to get focus' % id)


def wait_not_focused(id, timeout=10):
    global time_out_scale
    timeout *= time_out_scale
    DELAY = 0.5
    logger.debug('Waiting for %s window to lose focus...', id)
    for i in range(int(timeout/DELAY)):
        try:
            cur_id = xdotool(['getwindowfocus']).rstrip()
        except CalledProcessError:
            # When no window is available xdotool receives ID=1 and exits with error
            return
        logger.debug('Currently focused id: %s', cur_id)
        if cur_id != id:
            return
        time.sleep(DELAY)
    debug_window(cur_id)  # pragma: no cover
    raise RuntimeError('Timed out waiting for %s window to lose focus' % id)


def search_visible_windows(regex, others=None):
    """ Workaround for problems in xdotool failing to match window names """
    r = re.compile(regex)
    others = [] if others is None else others
    others_rx = [re.compile(v) for v in others]
    found = []
    windows = get_windows()
    # First check if we have one of the "others"
    for c, rx in enumerate(others_rx):
        for i in windows:
            if rx.search(i[1]):
                # Yes, inform it
                # Note: The main window can be focused with a dialog over it.
                #       If we found one of these dialogs it means we have a problem, no matters if the main windows is focused
                raise ValueError(others[c])
    # Now check for the window we need
    for i in windows:
        if r.search(i[1]):
            found.append(i[0])
    return found


def wait_for_window(name, window_regex, timeout=10, focus=True, skip_id=0, others=None, popen_obj=None):
    global time_out_scale
    timeout *= time_out_scale
    DELAY = 0.5
    logger.info('Waiting for "%s" ...', name)
    if skip_id:
        logger.debug('Will skip %s', skip_id)

    for i in range(int(timeout/DELAY)):
        try:
            window_id = search_visible_windows(window_regex, others=others)
            if len(window_id):
                logger.debug('Found %s window (%d)', name, len(window_id))
                if len(window_id) == 1:
                    id = window_id[0]
                if len(window_id) > 1:
                    id = window_id[1]
                logger.debug('Window id: %s', id)
                if id != skip_id:
                    if focus:
                        xdotool(['windowfocus', '--sync', id])
                        wait_focused(id, timeout)
                    return window_id
                else:
                    logger.debug('Skipped')
        except CalledProcessError:
            if popen_obj and popen_obj.poll() is not None:
                raise
        # Check if we have a list of alternative windows
        if others:
            for other in others:
                window_id = search_visible_windows(other)
                if len(window_id):
                    raise ValueError(other)
        if popen_obj:
            # Is KiCad running?
            ret_code = popen_obj.poll()
            if ret_code is not None:
                raise CalledProcessError(ret_code, 'KiCad')
        time.sleep(DELAY)
    debug_window()  # pragma: no cover
    raise RuntimeError('Timed out waiting for %s window' % name)


def wait_point(cfg):
    if cfg.wait_for_key:
        input('Press a key')


def capture_window_region(window_id, x, y, w, h, name):
    """ Capture a region of a window to a file """
    geometry = '{}x{}+{}+{}'.format(w, h, x, y)
    logger.debug('Capturing region {} from window {}'.format(geometry, window_id))
    name = os.path.join(img_tmp_dir, name)
    if not shutil.which('import'):
        logger.error("import isn't installed, please install it.\nThis is part of ImageMagick and GraphicsMagic packages.")
        sys.exit(MISSING_TOOL)
    res = check_output(['import', '-window', str(window_id), '-crop', geometry, name], stderr=DEVNULL).decode()
    logger.debug('Import output: ' + res)


def wait_window_get_ref(window_id, x, y, w, h):
    """ Takes a region of a window as reference image """
    global img_tmp_dir
    img_tmp_dir = mkdtemp(prefix='tmp-kiauto-images-')
    capture_window_region(window_id, x, y, w, h, "wait_ref.png")


def wait_window_change(window_id, x, y, w, h, time_out):
    """ Waits for a change in a window region """
    for i in range(int(time_out + 0.9)):
        capture_window_region(window_id, x, y, w, h, "current.png")
        current = os.path.join(img_tmp_dir, "current.png")
        wait_ref = os.path.join(img_tmp_dir, "wait_ref.png")
        difference = os.path.join(img_tmp_dir, "difference.png")
        res = run(['compare', '-fuzz', '5%', '-metric', 'AE', current, wait_ref, difference],
                  stderr=PIPE).stderr.decode()
        ae = int(res)
        logger.debug('Difference ' + res)
        if ae:
            shutil.rmtree(img_tmp_dir)
            return
        time.sleep(1)
    shutil.rmtree(img_tmp_dir)


def open_dialog_with_retry(msg, keys, desc, w_name, cfg):
    logger.info(msg)
    wait_point(cfg)
    if cfg.kicad_version >= KICAD_VERSION_5_99:
        # KiCad 6 has a very slow start-up
        time.sleep(1)
    xdotool(keys)
    retry = False
    try:
        id = wait_for_window(desc, w_name, popen_obj=cfg.popen_obj)
    except RuntimeError:  # pragma: no cover
        # Perhaps the main window wasn't available yet
        retry = True
    except CalledProcessError as e:
        logger.error(str(e))
        sys.exit(KICAD_DIED)
    if retry:
        logger.info('"{}" did not open, retrying'.format(desc))
        # wait_eeschema_start(cfg)
        xdotool(keys)
        try:
            id = wait_for_window(desc, w_name, popen_obj=cfg.popen_obj)
        except CalledProcessError as e:
            logger.error(str(e))
            sys.exit(KICAD_DIED)
    return id



# -*- coding: utf-8 -*-
# Copyright (c) 2020-2022 Salvador E. Tropea
# Copyright (c) 2020-2022 Instituto Nacional de Tecnologïa Industrial
# Copyright (c) 2019 Jesse Vincent (@obra)
# Copyright (c) 2018-2019 Seppe Stas (@seppestas) (Productize SPRL)
# Based on ideas by: Scott Bezek (@scottbez1)
# License: Apache 2.0
# Project: KiAuto (formerly kicad-automation-scripts)
# Adapted from: https://github.com/obra/kicad-automation-scripts
"""
Utilities related to the filesystem and file names.

Note: Only wait_for_file_created_by_process is from the original project.
"""
import os
import time
import re
import shutil
import atexit
from subprocess import DEVNULL
# python3-psutil
import psutil
import json

from kiauto.misc import (WRONG_ARGUMENTS, KICAD_VERSION_5_99, Config, READ_ONLY_PROBLEM)
from kiauto import log
logger = log.get_logger(__name__)
time_out_scale = 1.0


def set_time_out_scale(scale):
    global time_out_scale
    time_out_scale = scale


def wait_for_file_created_by_process(pid, file):
    global time_out_scale
    timeout = 15*time_out_scale
    process = psutil.Process(pid)
    DELAY = 0.2
    for j in range(2):
        # 2 passes: 1) for the file to be created 2) extend the timeout if created
        logger.debug('Waiting for file %s (pid %d) (timeout: %f)', file, pid, timeout)
        for i in range(int(timeout/DELAY)):
            kicad_died = False
            try:
                open_files = process.open_files()
            except psutil.AccessDenied:
                # Is our child, this access denied is because we are listing
                # files for other process that took the pid of the old KiCad.
                kicad_died = True
            if kicad_died:
                raise RuntimeError('KiCad unexpectedly died')
            logger.debug(open_files)
            if os.path.isfile(file):
                file_open = False
                for open_file in open_files:
                    if open_file.path == file:
                        file_open = True
                if file_open:
                    logger.debug('Waiting for process to close file')
                else:
                    return
            else:
                logger.debug('Waiting for process to create file')
            time.sleep(DELAY)
        # If the file was created assume KiCad is working
        if os.path.isfile(file):
            timeout = 45*time_out_scale
        else:
            # The file wasn't even created, don't wait much
            timeout = 1*time_out_scale

    raise RuntimeError('Timed out waiting for creation of %s' % file)


def load_filters(cfg, file):
    """ Load errors filters """
    if not os.path.isfile(file):
        logger.error("Filter file `{}` doesn't exist".format(file))
        exit(WRONG_ARGUMENTS)
    logger.debug('Loading filter errors')
    with open(file, 'r') as f:
        ln = 1
        fl = 0
        for line in f:
            line = line.rstrip()
            if len(line) > 0 and line[0] != '#':
                m = re.search(r'^(\S+)\s*,(.*)$', line)
                if m:
                    cfg.err_filters.append([m.group(1), m.group(2)])
                    fl = fl+1
                else:
                    logger.error('Syntax error at line {} in filter file `{}`: `{}`'.format(ln, file, line))
                    logger.error('Use `ERROR_NUMBER,REGEX` format')
                    exit(WRONG_ARGUMENTS)
            ln = ln+1
        logger.info('Loaded {} error filters from `{}`'.format(fl, file))


def apply_filters(cfg, err_name, wrn_name):
    """ Apply the error filters to the list of errors and unconnecteds """
    if len(cfg.err_filters) == 0:
        return (0, 0)
    skip_err = 0
    for i, err in enumerate(cfg.errs):
        for f in cfg.err_filters:
            if err.startswith('({})'.format(f[0])):
                m = re.search(f[1], err)
                if m:
                    skip_err += 1
                    logger.warning('Ignoring '+err)
                    logger.debug('Matched regex `{}`'.format(f[1]))
                    cfg.errs[i] = None
                    break
    if skip_err:
        logger.info('Ignoring {} {}'.format(skip_err, err_name))
    skip_wrn = 0
    for i, wrn in enumerate(cfg.wrns):
        for f in cfg.err_filters:
            if wrn.startswith('({})'.format(f[0])):
                m = re.search(f[1], wrn)
                if m:
                    skip_wrn += 1
                    logger.info('Ignoring '+wrn)
                    logger.debug('Matched regex `{}`'.format(f[1]))
                    cfg.wrns[i] = None
                    break
    if skip_wrn:
        logger.info('Ignoring {} {}'.format(skip_wrn, wrn_name))
    return skip_err, skip_wrn


def list_errors(cfg):
    for err in cfg.errs:
        if err:
            if "; Severity: warning" in err:
                logger.warning(err.replace("; Severity: warning", ''))
            else:
                logger.error(err.replace("; Severity: error", ''))


def list_warnings(cfg):
    for wrn in cfg.wrns:
        if wrn:
            if "; Severity: error" in wrn:
                logger.error(wrn.replace("; Severity: error", ''))
            else:
                logger.warning(wrn.replace("; Severity: warning", ''))


def check_kicad_config_dir(cfg):
    if not os.path.isdir(cfg.kicad_conf_path):
        logger.debug('Creating KiCad config dir')
        os.makedirs(cfg.kicad_conf_path, exist_ok=True)


def remove_lib_table(fname):
    if os.path.isfile(fname):
        logger.debug('Removing '+fname)
        os.remove(fname)


def check_lib_table(fuser, fsys):
    if not os.path.isfile(fuser):
        logger.debug('Missing default sym-lib-table')
        for f in fsys:
            if os.path.isfile(f):
                shutil.copy2(f, fuser)
                return
        logger.warning('Missing default system symbol table '+fsys[0] +
                       ' creating an empty one')  # pragma: no cover
        with open(fuser, 'wt') as f:
            f.write('({} )\n'.format(os.path.basename(fuser).replace('-', '_')))
        atexit.register(remove_lib_table, fuser)


def restore_one_config(name, fname, fbkp):
    if fbkp and os.path.exists(fbkp):
        if os.path.exists(fname):
            os.remove(fname)
        os.rename(fbkp, fname)
        logger.debug('Restoring old %s config', name)
        return None
    return fbkp


def restore_config(cfg):
    """ Restore original user configuration """
    cfg.conf_eeschema_bkp = restore_one_config('eeschema', cfg.conf_eeschema, cfg.conf_eeschema_bkp)
    cfg.conf_kicad_bkp = restore_one_config('KiCad common', cfg.conf_kicad, cfg.conf_kicad_bkp)
    cfg.conf_hotkeys_bkp = restore_one_config('user hotkeys', cfg.conf_hotkeys, cfg.conf_hotkeys_bkp)
    cfg.conf_pcbnew_bkp = restore_one_config('pcbnew', cfg.conf_pcbnew, cfg.conf_pcbnew_bkp)
    cfg.conf_colors_bkp = restore_one_config('colors', cfg.conf_colors, cfg.conf_colors_bkp)
    cfg.conf_3dview_bkp = restore_one_config('3D viewer', cfg.conf_3dview, cfg.conf_3dview_bkp)


def backup_config(name, file, err, cfg):
    config_file = file
    old_config_file = file+'.pre_script'
    logger.debug(name+' config: '+config_file)
    # If we have an old back-up ask for the user to solve it
    if os.path.isfile(old_config_file):
        logger.error(name+' config back-up found (%s)', old_config_file)
        logger.error('It could contain your %s configuration, rename it to %s or discard it.', name.lower(), config_file)
        exit(err)
    if os.path.isfile(config_file):
        logger.debug('Moving current config to '+old_config_file)
        os.rename(config_file, old_config_file)
        atexit.register(restore_config, cfg)
        return old_config_file
    return None


def create_user_hotkeys(cfg):
    logger.debug('Creating a user hotkeys config')
    with open(cfg.conf_hotkeys, "wt") as text_file:
        text_file.write('common.Control.print\tCtrl+P\n')
        text_file.write('common.Control.plot\tCtrl+Shift+P\n')
        text_file.write('common.Control.show3DViewer\tAlt+3\n')
        text_file.write('eeschema.EditorControl.exportNetlist\tCtrl+Shift+N\n')
        text_file.write('eeschema.EditorControl.generateBOM\tCtrl+Shift+B\n')
        text_file.write('eeschema.InspectionTool.runERC\tCtrl+Shift+I\n')
        text_file.write('pcbnew.DRCTool.runDRC\tCtrl+Shift+I\n')
        text_file.write('pcbnew.ZoneFiller.zoneFillAll\tB\n')
        text_file.write('pcbnew.EditorControl.generateD356File\tAlt+Shift+E\n')
        text_file.write('3DViewer.Control.rotateXclockwise\tAlt+X\n')
        text_file.write('3DViewer.Control.rotateXcounterclockwise\tAlt+Shift+X\n')
        text_file.write('3DViewer.Control.rotateYclockwise\tAlt+Y\n')
        text_file.write('3DViewer.Control.rotateYcounterclockwise\tAlt+Shift+Y\n')
        text_file.write('3DViewer.Control.rotateZclockwise\tAlt+Z\n')
        text_file.write('3DViewer.Control.rotateZcounterclockwise\tAlt+Shift+Z\n')


def create_kicad_config(cfg):
    logger.debug('Creating a KiCad common config')
    with open(cfg.conf_kicad, "wt") as text_file:
        if cfg.conf_kicad_json:
            kiconf = {"environment": {"show_warning_dialog": False}}
            kiconf['graphics'] = {"cairo_antialiasing_mode": 0, "opengl_antialiasing_mode": 0}
            kiconf['system'] = {"editor_name": "/bin/cat"}
            # Copy the environment vars if available
            if cfg.conf_kicad_bkp:
                vars = Config.get_config_vars_json(cfg.conf_kicad_bkp)
                if vars:
                    kiconf['environment']['vars'] = vars
            text_file.write(json.dumps(kiconf))
            logger.debug(json.dumps(kiconf))
        else:
            text_file.write('ShowEnvVarWarningDialog=0\n')
            text_file.write('Editor=/bin/cat\n')
            # Copy the environment vars if available
            if cfg.conf_kicad_bkp:
                vars = Config.get_config_vars_ini(cfg.conf_kicad_bkp)
                if vars:
                    text_file.write('[EnvironmentVariables]\n')
                    for key in vars:
                        text_file.write(key.upper()+'='+vars[key]+'\n')


def restore_autosave(name):
    """ Restores de auto save information """
    old_name = name[:-11]
    if os.path.isfile(name):
        logger.debug('Restoring {} -> {}'.format(name, old_name))
        os.rename(name, old_name)


def check_input_file(cfg, no_file, no_ext):
    # Check the schematic/PCB is there
    if not os.path.isfile(cfg.input_file):
        logger.error(cfg.input_file+' does not exist')
        exit(no_file)
    # If we pass a name without extension KiCad will try to create a kicad_sch/kicad_pcb
    # The extension can be anything.
    ext = os.path.splitext(cfg.input_file)[1]
    if not ext:
        logger.error('Input files must use an extension, otherwise KiCad will reject them.')
        exit(no_ext)
    if cfg.kicad_version >= KICAD_VERSION_5_99 and ext == '.sch':
        logger.warning('Using old format files is not recommended. Convert them first.')
    # KiCad 6 uses #auto_saved_files# to store autosave info
    fauto = os.path.join(os.path.dirname(cfg.input_file), '#auto_saved_files#')
    if os.path.isfile(fauto):
        logger.warning('Partially saved project detected, please double check it')
        # Rename it so KiCad doesn't ask about restoring autosaved files
        fauto_new = fauto+'.moved_away'
        logger.debug('Renaming {} -> {}'.format(fauto, fauto_new))
        try:
            os.rename(fauto, fauto_new)
        except PermissionError:
            # Read-only directory or file system, give up
            logger.error('Unable to rename `{}` please remove it manually'.format(fauto))
            exit(READ_ONLY_PROBLEM)
        # Restore it at exit
        atexit.register(restore_autosave, fauto_new)


def memorize_project(cfg):
    """ Detect the .pro filename and try to read it and its mtime.
        If KiCad changes it then we'll try to revert the changes """
    cfg.pro_stat = None
    cfg.pro_content = None
    cfg.prl_stat = None
    cfg.prl_content = None
    name_no_ext = os.path.splitext(cfg.input_file)[0]
    cfg.pro_name = name_no_ext+'.'+cfg.pro_ext
    if not os.path.isfile(cfg.pro_name):
        cfg.pro_name = name_no_ext+'.pro'
        if not os.path.isfile(cfg.pro_name):
            logger.warning('KiCad project file not found')
            cfg.pro_name = name_no_ext+'.'+cfg.pro_ext
            return
        if cfg.kicad_version >= KICAD_VERSION_5_99:
            logger.warning('Using old format projects is not recommended. Convert them first.')
    if cfg.pro_name[-4:] == '.pro':
        cfg.pro_stat = cfg.start_pro_stat
    else:
        cfg.pro_stat = cfg.start_kicad_pro_stat
    with open(cfg.pro_name) as f:
        cfg.pro_content = f.read()
    atexit.register(restore_project, cfg)
    if cfg.prl_ext:
        cfg.prl_name = name_no_ext+'.'+cfg.prl_ext
        if not os.path.isfile(cfg.prl_name):
            return
        cfg.prl_stat = cfg.start_kicad_prl_stat
        with open(cfg.prl_name) as f:
            cfg.prl_content = f.read()


def _restore_project(name, stat_v, content):
    logger.debug('Checking if %s was modified', name)
    if stat_v and content:
        pro_found = False
        if os.path.isfile(name):
            new_stat = os.stat(name)
            pro_found = True
        else:  # pragma: no cover
            logger.warning('Project file lost')
        if not pro_found or new_stat.st_mtime != stat_v.st_mtime:
            logger.debug('Restoring the project file')
            os.rename(name, name+'-bak')
            with open(name, 'wt') as f:
                f.write(content)
            os.utime(name, times=(stat_v.st_atime, stat_v.st_mtime))


def restore_project(cfg):
    """ If the .pro was modified try to restore it """
    _restore_project(cfg.pro_name, cfg.pro_stat, cfg.pro_content)
    if cfg.prl_ext and cfg.prl_stat:
        _restore_project(cfg.prl_name, cfg.prl_stat, cfg.prl_content)


def get_log_files(out_dir, app_name):
    if log.get_level() > 2:
        flog_out = open(os.path.join(out_dir, app_name+'_out.log'), 'wt')
        flog_err = open(os.path.join(out_dir, app_name+'_err.log'), 'wt')
        logger.debug('Redirecting '+app_name+' output to '+app_name+'*.log')
    else:
        flog_out = flog_err = DEVNULL
    return (flog_out, flog_err)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Salvador E. Tropea
# Copyright (c) 2020 Instituto Nacional de Tecnologïa Industrial
# License: Apache 2.0
# Project: KiAuto (formerly kicad-automation-scripts)
# Adapted from: https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
"""
Log module

Handles logging initialization and formating.
"""
import sys
import logging

# Default domain, base name for the tool
domain = 'kiauto'
verbose_level = 2


def get_logger(name=None):
    """Get a module for a submodule or the root logger if no name is provided"""
    return logging.getLogger(domain+'.'+name) if name else logging.getLogger(domain)


def set_domain(name):
    """Set the base name for the tool"""
    global domain
    domain = name


def set_level(logger, level):
    global verbose_level
    verbose_level = level
    if level >= 2:
        log_level = logging.DEBUG
    elif level == 1:
        log_level = logging.INFO
    else:
        log_level = logging.WARNING
    logger.setLevel(log_level)


def get_level():
    return verbose_level


def init():
    """Initialize the logging feature using a custom format and the specified verbosity level"""
    logger = get_logger()
    ch = logging.StreamHandler()
    ch.setFormatter(CustomFormatter())
    logger.addHandler(ch)
    return logger


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors"""

    if sys.stderr.isatty():
        grey = "\x1b[38;21m"
        yellow = "\x1b[93;1m"
        red = "\x1b[91;1m"
        bold_red = "\x1b[91;21m"
        cyan = "\x1b[36;1m"
        reset = "\x1b[0m"
    else:
        grey = ""
        yellow = ""
        red = ""
        bold_red = ""
        cyan = ""
        reset = ""
    # format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"
    format = "%(levelname)s:%(message)s (%(name)s - %(filename)s:%(lineno)d)"
    format_simple = "%(levelname)s:%(message)s"

    FORMATS = {
        logging.DEBUG: cyan + format + reset,
        logging.INFO: grey + format_simple + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)
